/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include <fggl/gui/containers.hpp>
#include <spdlog/spdlog.h>

namespace fggl::gui {

	Widget *Container::getChildAt(const math::vec2 &point) {
		for (auto &child : m_children) {
			if (child->contains(point)) {
				return child->getChildAt(point);
			}
		}

		return nullptr;
	}

	bool Container::contains(const math::vec2 &point) {
		return true;
	}

	void Container::render(gfx::Paint &paint) {
		for (auto &child : m_children) {
			child->render(paint);
		}
	}

	void Container::add(std::unique_ptr<Widget> widget) {
		m_children.push_back(std::move(widget));
		m_dirty = true;
	}

	void Container::onMouseOver(math::vec2 pos) {
		Widget::onMouseOver(pos);
		auto* childHover = getChildAt(pos);

		for ( auto& child : m_children) {
			if ( child.get() != childHover ) {
				child->onExit(pos);
			}
		}

		if ( childHover != nullptr ) {
			childHover->onMouseOver(pos);
		}
	}

	void Container::onEnter(math::vec2i pos) {
		Widget::onEnter(pos);
		for ( auto& child : m_children ) {
			if ( child->contains(pos)) {
				child->onEnter(pos);
			}
		}
	}

	void Container::onExit(math::vec2i pos) {
		Widget::onExit(pos);
		for ( auto& child : m_children) {
			child->onExit(pos);
		}
	}

	GridBox::GridBox(uint32_t rows, uint32_t cols, uint32_t padx, uint32_t pady) : m_rows(rows), m_cols(cols), m_padding(padx, pady) {}

	void GridBox::layout() {
		assert( m_rows != 0 || m_cols != 0 );

		if ( m_rows == 0 ) {
			int rows = m_children.size() / m_cols;

			// figure out the width and heights
			float* widths = new float[m_cols]{0.0F};
			float* heights = new float[rows]{0.0F};
			for ( auto idx = 0U; idx < m_children.size(); ++idx) {
				auto& child = m_children[idx];
				int col = idx % m_cols;
				int row = idx / m_cols;

				widths[col] = std::max( child->size().x, widths[col] );
				heights[row] = std::max( child->size().y, heights[row] );
			}

			// populate the grid
			fggl::math::vec2i pos{0, 0};
			int row = 0;
			int col = 0;
			for ( auto& child : m_children ) {
				fggl::math::vec2i size{ widths[col], heights[row] };
				child->size(pos, size);
				child->layout();

				// next iter
				pos.x += size.x + m_padding.x;
				col++;
				if ( col == m_cols ) {
					col = 0;
					row++;
					pos.x = 0;
					pos.y += size.y + m_padding.y;
				}
			}

		}
	}


	/*
	Box::Box( LayoutAxis axis ) : m_axis( axis ) {}
	
	void Box::layout() {
		math::vec2 baseSize{};
		
		const int mainAxis = 0;
		const int crossAxis = 1;
		
		// ask for preferred sizes
		for ( auto& child : m_children ) {
			auto childBaseSize = child.baseSize();
			
			baseSize[mainAxis] += childBaseSize[mainAxis];
			if ( childBaseSize[crossAxis] > baseSize[crossAxis] ) {
				baseSize[crossAxis] = childBaseSize[crossAxis];			
			}
		}
		
		if ( lineAxis == 0 )
			setSize(lineSum, biggestCross);
		else
			setSize(biggestCross, lineSum);
	}*/

	void Panel::render(gfx::Paint &paint) {
		// background painting time
		gfx::Path2D background(topLeft());
		background.colour(math::vec3(32.0f / 255.0f, 74.0F / 255.0F, 135.0F / 255.0F));
		draw_box(background, topLeft(), bottomRight());
		paint.fill(background);

		// call parent's render method for child objects
		Container::render(paint);
	}
};
