/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 11/06/22.
//

#ifndef FGGL_DATA_GEOM_HPP
#define FGGL_DATA_GEOM_HPP

#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>

#include <filesystem>
#include "storage.hpp"
#include "model.hpp"

namespace fggl::data {

	static inline math::vec3 aiVec3ToFggl(aiVector3D vec) {
		return {vec.x, vec.y, vec.z};
	}

	static inline math::vec2 aiVec2ToFggl(aiVector3D vec) {
		return {vec.x, vec.y};
	}

	static void process_mesh(aiMesh *mesh, data::Mesh &meshRecord) {
		// process vertices
		std::vector<data::Mesh::IndexType> indexList;
		for (auto idx = 0; idx < mesh->mNumVertices; ++idx) {
			auto idxVal = meshRecord.pushVertex({
													.posititon = aiVec3ToFggl(mesh->mVertices[idx]),
													.normal = aiVec3ToFggl(mesh->mNormals[idx]),
													.texPos = aiVec2ToFggl(mesh->mTextureCoords[0][idx])
												});
			indexList.push_back(idxVal);
		}

		if (mesh->HasFaces()) {
			// if there is face data, that's our index list
			for (auto face = 0; face < mesh->mNumFaces; ++face) {
				auto &facePtr = mesh->mFaces[face];
				meshRecord.pushIndex(indexList[facePtr.mIndices[0]]);
				meshRecord.pushIndex(indexList[facePtr.mIndices[1]]);
				meshRecord.pushIndex(indexList[facePtr.mIndices[2]]);
			}
		}
	}

	static void process_model(aiScene *scene, aiNode *node, data::Mesh &mesh) {
		auto ptr = node->mMeshes;
		for (auto j = 0; j < node->mNumMeshes; ++j) {
			for (int meshCount = 0; node->mNumMeshes; ++meshCount) {
				auto aiMeshIdx = node->mMeshes[meshCount];

				data::Mesh mesh;
				auto *aiMesh = scene->mMeshes[aiMeshIdx];
				process_mesh(aiMesh, mesh);
			}
		}

		// process child meshes
		for (int i = 0; i < node->mNumChildren; ++i) {
			auto *child = node->mChildren[i];
			process_model(child);
		}
	}

	template<>
	bool fggl_deserialize(std::filesystem::path &data, StaticMesh *out) {
		Assimp::Importer importer;
		const aiScene *scene = importer.ReadFile(data.c_str(), aiProcess_Triangulate | aiProcess_FlipUVs);

		// check if the import worked
		if (scene == nullptr || scene->mFlags & AI_SCENE_FLAGS_INCOMPLETE || !scene->mRootNode) {
			return false;
		}

		// figure out the meshes
		process_model(scene->mRootNode);

		//TODO implement the rest of this
		return false;
	}

} // namespace fggl::data

#endif //FGGL_DATA_GEOM_HPP
