/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 24/07/22.
//

#ifndef FGGL_ENTITY_LOADER_SERIALISE_HPP
#define FGGL_ENTITY_LOADER_SERIALISE_HPP

#include "yaml-cpp/yaml.h"

#include "fggl/math/types.hpp"
#include "fggl/data/model.hpp"
#include "fggl/phys/types.hpp"

namespace YAML {

	template<>
	struct convert<fggl::math::vec3> {
		static Node encode(const fggl::math::vec3 &rhs) {
			Node node;
			node.push_back(rhs.x);
			node.push_back(rhs.y);
			node.push_back(rhs.z);
			return node;
		}

		static bool decode(const Node &node, fggl::math::vec3 &rhs) {
			if (!node.IsSequence() || node.size() != 3) {
				return false;
			}

			rhs.x = node[0].as<float>();
			rhs.y = node[1].as<float>();
			rhs.z = node[2].as<float>();
			return true;
		}
	};

	template<>
	struct convert<fggl::math::vec2> {
		static Node encode(const fggl::math::vec2 &rhs) {
			Node node;
			node.push_back(rhs.x);
			node.push_back(rhs.y);
			return node;
		}

		static bool decode(const Node &node, fggl::math::vec2 &rhs) {
			if (!node.IsSequence() || node.size() != 2) {
				return false;
			}

			rhs.x = node[0].as<float>();
			rhs.y = node[1].as<float>();
			return true;
		}
	};

	template<>
	struct convert<fggl::data::Vertex> {
		static Node encode(const fggl::data::Vertex &rhs) {
			Node node;
			node["position"] = rhs.posititon;
			node["normal"] = rhs.normal;
			node["colour"] = rhs.colour;
			node["texPos"] = rhs.texPos;
			return node;
		}

		static bool decode(const Node &node, fggl::data::Vertex &rhs) {
			if (!node.IsSequence() || node.size() != 2) {
				return false;
			}

			rhs.posititon = node["position"].as<fggl::math::vec3>();
			rhs.normal = node["normal"].as<fggl::math::vec3>();
			rhs.colour = node["colour"].as<fggl::math::vec3>();
			rhs.texPos = node["texPos"].as<fggl::math::vec2>();
			return true;
		}
	};

	template<>
	struct convert<fggl::phys::BodyType> {
		static Node encode(const fggl::phys::BodyType &rhs) {
			Node node;
			if (rhs == fggl::phys::BodyType::STATIC) {
				node = "static";
			} else if (rhs == fggl::phys::BodyType::DYNAMIC) {
				node = "dynamic";
			} else if (rhs == fggl::phys::BodyType::KINEMATIC) {
				node = "kinematic";
			}
			return node;
		}

		static bool decode(const Node &node, fggl::phys::BodyType &rhs) {
			auto strVal = node.as<std::string>();
			if (strVal == "static") {
				rhs = fggl::phys::BodyType::STATIC;
				return true;
			}

			if (strVal == "dynamic") {
				rhs = fggl::phys::BodyType::DYNAMIC;
				return true;
			}

			if (strVal == "kinematic") {
				rhs = fggl::phys::BodyType::KINEMATIC;
				return true;
			}

			return false;
		}
	};

	template<>
	struct convert<fggl::util::GUID> {
		static Node encode(const fggl::util::GUID &rhs) {
			Node node;
			node = rhs.get();
			return node;
		}

		static bool decode(const Node &node, fggl::util::GUID &rhs) {
			auto longVal = node.as<uint64_t>(0);

			if (longVal == 0) {
				// probably meant to hash it...
				auto stringVal = node.as<std::string>();
				rhs = fggl::util::make_guid_rt(stringVal);
				return true;
			}

			// it's probably pre-hashed...
			rhs = fggl::util::GUID::make(longVal);
			return true;
		}
	};

}

#endif //FGGL_ENTITY_LOADER_SERIALISE_HPP
