/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/06/22.
//

#ifndef FGGL_UTIL_SAFETY_HPP
#define FGGL_UTIL_SAFETY_HPP

#include <string>

namespace fggl::util {

	/**
	 * A type-safe opaque handle.
	 *
	 * Lots of low-level libraries we use pass around handles as some primative type. It's fairly easy to accidentally
	 * mix these up. This wrapper's job is to make sure that mixing up handle types is impossible (and results in
	 * compiler errors).
	 *
	 * @tparam T the underling type of the handle.
	 * @tparam Tag a unique tag used to identify the handle type.
	 */
	template<typename T, typename Tag>
	struct OpaqueName {
		private:
			T m_value;

		public:
			explicit constexpr OpaqueName(T value) : m_value(value) {}

			constexpr OpaqueName() : m_value() {}

			constexpr T get() const {
				return m_value;
			}

			constexpr explicit operator std::string_view() const { return m_value; }

			/**
			 * Check for equality of two handles.
			 *
			 * Two values are considered the same of the values contained inside them are considered equal, and both
			 * types share the same tagging interface.
			 *
			 * @param other the value being compared against.
			 * @return true iff the contained values are equal
			 */
			bool operator==(const OpaqueName<T, Tag> &other) const {
				return m_value == other.m_value;
			}

			/**
			 * Check for equality of two handles.
			 *
			 * Two values are considered the same of the values contained inside them are considered equal, and both
			 * types share the same tagging interface.
			 *
			 * @param other the value being compared against.
			 * @return true iff the contained values are not equal
			 */
			bool operator!=(const OpaqueName<T, Tag> &other) const {
				return m_value != other.m_value;
			}

			bool operator<(const OpaqueName<T, Tag> &other) const {
				return m_value < other.m_value;
			}

			std::strong_ordering operator<=>(const OpaqueName<T, Tag> &other) const noexcept {
				return m_value <=> other.m_value;
			}

			/**
			 * Generate a new tagged instance of a handle.
			 */
			constexpr static OpaqueName<T, Tag> make(T value) {
				return OpaqueName<T, Tag>(value);
			}
	};

} // namespace fggl::util

#endif //FGGL_UTIL_SAFETY_HPP
